//
// period_locker.c
//
#include "period_locker.h"

#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <pthread.h>
#include <semaphore.h>
#include <time.h>
#include <unistd.h>
#include <sys/file.h>
#include <sys/mman.h>


#ifndef ENABLE_TRACE
//# define ENABLE_TRACE	1
# define ENABLE_TRACE	0
#endif

const uint32_t	SLEEP_UNIT = 100;  // 100[ms]
const char	LOCK_NAME[] = "/dev/shm/se050_lock";
const char	SHMEM_NAME[] = "/se050_usage";

typedef struct se050_usage {
	int				num_user;
} se050_usage;

typedef struct period_locker {
	int				shared_lock;
	bool			has_lock;
	bool			is_release_reserved;
	uint64_t		lock_time;
	uint64_t		release_rtime;
	pthread_t		release_thread;
	pthread_mutex_t	inner_lock;
	se050_usage*	usage_info;
	bool			quit_requested;
} period_locker;


static uint64_t
GetCurrentTimeInMillis()
{
	struct timespec	ts;

	(void)clock_gettime(CLOCK_MONOTONIC_RAW, &ts);

	return (ts.tv_sec * 1000 + ts.tv_nsec / 1000000);
}

static void
SleepMillis(uint32_t msec)
{
	usleep(msec * 1000);
}

static bool
TimedWaitForLock(int lock_fd, uint32_t timeoutMsec)
{
	uint64_t	startTime = GetCurrentTimeInMillis();

	if (0 == flock(lock_fd, LOCK_EX | LOCK_NB)) {
		return true;
	} else if (EWOULDBLOCK != errno) {
		return false;
	} else {
		do {
			SleepMillis(SLEEP_UNIT);
			if (0 == flock(lock_fd, LOCK_EX | LOCK_NB)) {
				return true;
			} else if (EWOULDBLOCK != errno) {
				return false;
			}
		} while (GetCurrentTimeInMillis() - startTime < timeoutMsec);
		return false;  // timed out
	}
}

static se050_usage*
CreateOrOpen_UsageSharedMem()
{
	se050_usage*	shm = NULL;
	bool	created = false;
	int	fd = shm_open(SHMEM_NAME, O_RDWR, 0666);

	if (fd < 0 && ENOENT == errno) {
		fd = shm_open(SHMEM_NAME, O_CREAT|O_RDWR, 0666);
		if (fd >= 0) {
			if (0 != ftruncate(fd, sizeof(se050_usage))) {
				close(fd);
				return NULL;
			}
			created = true;
		}
	}
	if (fd < 0) {
		return NULL;
	}

	shm = mmap(NULL, sizeof(se050_usage), PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
	close(fd);
	if (NULL == shm) {
		return NULL;
	}
#if ENABLE_TRACE
	if (! created) {
		printf("shared mem is opend, user count= %d\n", shm->num_user);
	}
#endif  // ENABLE_TRACE

	return shm;
}

static void
Close_UsageSharedMem(se050_usage* shm, bool doFree)
{
	if (doFree) {
		(void)shm_unlink(SHMEM_NAME);
	}
	munmap(shm, sizeof(*shm));
}

static void
PeriodLocker_DoReleaseLock(period_locker* obj)
{
	obj->lock_time = 0;
	obj->has_lock = obj->is_release_reserved = false;
	(void)flock(obj->shared_lock, LOCK_UN);
}

static void*
PeriodLocker_ThreadProc(void* arg)
{
	period_locker* obj = (period_locker*)arg;

	// assertion
	if (! obj->has_lock) {
		fprintf(stderr, "Invalid argument for %s().\n", __FUNCTION__);
		return (void*)-1;
	}

	obj->quit_requested = false;
	for (;;) {
		uint64_t	currTime;

		(void)pthread_mutex_lock(&obj->inner_lock);
		if (obj->quit_requested) {
			obj->release_thread = 0;
			obj->release_rtime  = 0;
			(void)pthread_mutex_unlock(&obj->inner_lock);
			break;
		}
		currTime = GetCurrentTimeInMillis();
		if (obj->has_lock) {
			if (obj->is_release_reserved) {
				if (currTime - obj->release_rtime >= LOCK_PERIOD) {
#if ENABLE_TRACE
					printf("reached to the period, release the lock.\n");
#endif  // ENABLE_TRACE
					PeriodLocker_DoReleaseLock(obj);
				}
			} else if (currTime - obj->lock_time >= MAX_LOCK_PERIOD) {
				fprintf(stderr, "the lock is held too long without update, so release it.\n");
				PeriodLocker_DoReleaseLock(obj);
				obj->release_thread = 0;
				obj->release_rtime  = 0;
				(void)pthread_mutex_unlock(&obj->inner_lock);
				break;
			}
		}
		if ((0 != obj->release_rtime) && (currTime - obj->release_rtime >= READY_PERIOD)) {
#if ENABLE_TRACE
			printf("expires the living period, finish the thread.\n");
#endif  // ENABLE_TRACE
			obj->release_thread = 0;
			obj->release_rtime  = 0;
			(void)pthread_mutex_unlock(&obj->inner_lock);
			break;
		}
		(void)pthread_mutex_unlock(&obj->inner_lock);

		if (! obj->quit_requested) {
			SleepMillis(SLEEP_UNIT);
		}
	}

	return 0;
}


period_locker_t
PeriodLocker_Create()
{
	period_locker*	newObj = malloc(sizeof(period_locker));
	int lock_fd;

	if (NULL == newObj) {
		fprintf(stderr, "failed to malloc().\n");
		return NULL;
	}
	memset(newObj, 0, sizeof(*newObj));
	lock_fd = open(LOCK_NAME, O_CREAT | O_RDWR, 00666);
	if (lock_fd < 0) {
		fprintf(stderr, "failed to oopen the lock file.\n");
		free(newObj);
		return NULL;
	}
	newObj->shared_lock = lock_fd;

	(void)pthread_mutex_init(&newObj->inner_lock, NULL);
	if (! PeriodLocker_AcquireLock((period_locker_t)newObj)) {
		fprintf(stderr, "failed to PeriodLocker_AcquireLock().\n");
		(void)pthread_mutex_destroy(&newObj->inner_lock);
		(void)close(newObj->shared_lock);
		free(newObj);
		return NULL;
	}
	newObj->usage_info = CreateOrOpen_UsageSharedMem();
	if (NULL == newObj->usage_info) {
		fprintf(stderr, "failed to Create_UsageSharedMem().\n");
		PeriodLocker_DoReleaseLock(newObj);
		(void)pthread_mutex_destroy(&newObj->inner_lock);
		(void)close(newObj->shared_lock);
		free(newObj);
		return NULL;
	}
	newObj->usage_info->num_user += 1;
#if ENABLE_TRACE
	printf("newObj->usage_info->num_user= %d\n", newObj->usage_info->num_user);
#endif  // ENABLE_TRACE
	PeriodLocker_DoReleaseLock(newObj);

	return (period_locker_t)newObj;
}

void
PeriodLocker_Destroy(period_locker_t locker,
	void(*shared_res_free_callback(void*)), void* cbArg)
{
	period_locker*	obj = (period_locker*)locker;

	(void)pthread_mutex_lock(&obj->inner_lock);
	if (0 != obj->release_thread) {
		obj->quit_requested = true;
		(void)pthread_mutex_unlock(&obj->inner_lock);
		while (0 != obj->release_thread) {
			SleepMillis(10);
		}
		(void)pthread_mutex_lock(&obj->inner_lock);
		if (obj->has_lock) {
			PeriodLocker_DoReleaseLock(obj);
		}
	}
	(void)pthread_mutex_unlock(&obj->inner_lock);

	if (! obj->has_lock) {
		if (! TimedWaitForLock(obj->shared_lock, 30 * 1000)) {
			fprintf(stderr, "failed to acquire the lock.\n");
			Close_UsageSharedMem(obj->usage_info, false);
			(void)pthread_mutex_destroy(&obj->inner_lock);
			(void)close(obj->shared_lock);
			return;
		}
		obj->has_lock = true;
	}
	obj->usage_info->num_user -= 1;
	if (obj->usage_info->num_user == 0) {
#if ENABLE_TRACE
		printf("the reference count reaches 0, so release shared resources.\n");
#endif  // ENABLE_TRACE
		if (NULL != shared_res_free_callback) {
			shared_res_free_callback(cbArg);
		}
		Close_UsageSharedMem(obj->usage_info, true);
	} else {
		Close_UsageSharedMem(obj->usage_info, false);
	}
	obj->usage_info = NULL;
	PeriodLocker_DoReleaseLock(obj);

	(void)pthread_mutex_destroy(&obj->inner_lock);
	(void)close(obj->shared_lock);
	free(obj);
//
// Note: When the obj->usage_info->num_user becomse 0, it must to 
//       release the session with SE050 before releasing the lock.
//       There fore, it must take a pointer to the function that performs 
//       session release processing as a callback argument, and call it 
//       before reelasing the lock.
//       (2023-11-16, shinya.koga@atmark-techno.com)
}

bool
PeriodLocker_AcquireLock(period_locker_t locker)
{
	period_locker*	obj = (period_locker*)locker;
	se050_usage*	usage_info = obj->usage_info;

	(void)pthread_mutex_lock(&obj->inner_lock);
	obj->is_release_reserved = false;
	if (! obj->has_lock) {

		if (0 == flock(obj->shared_lock, LOCK_EX|LOCK_NB)) {
			obj->has_lock = true;
#if ENABLE_TRACE
			printf("the lock is acquired immediately\n");
#endif  // ENABLE_TRACE
		} else if (EWOULDBLOCK == errno) {
			if (NULL == usage_info) {
				// in case of called from PeriodLocker_Create()
				usage_info = obj->usage_info = CreateOrOpen_UsageSharedMem();
				if (NULL == usage_info) {
					fprintf(stderr, "failed to CreateOrOpen_UsageSharedMem().\n");
					goto finish;
				}
			}
			if (TimedWaitForLock(obj->shared_lock, 30 * 1000)) {
				obj->has_lock = true;
#if ENABLE_TRACE
				printf("the lock is acquired after waiting.\n");
#endif  // ENABLE_TRACE
			} else {
				fprintf(stderr, "failed to TimedWaitForLock().\n");
				(void)pthread_mutex_unlock(&obj->inner_lock);
				return false;
			}
		} else {
			fprintf(stderr, "failed to trying the lock.\n");
			(void)pthread_mutex_unlock(&obj->inner_lock);
			return false;
		}

		if (0 == obj->release_thread) {
			pthread_attr_t	thread_attr;

			(void)pthread_attr_init(&thread_attr);
			(void)pthread_attr_setdetachstate(&thread_attr, PTHREAD_CREATE_DETACHED);
			if (0 != pthread_create(&obj->release_thread, &thread_attr, PeriodLocker_ThreadProc, obj)) {
				fprintf(stderr, "failed to pthread_create().\n");
				PeriodLocker_DoReleaseLock(obj);
				(void)pthread_attr_destroy(&thread_attr);
				goto finish;
			}
			(void)pthread_attr_destroy(&thread_attr);
		}
	}
	obj->lock_time = GetCurrentTimeInMillis();
finish:
	(void)pthread_mutex_unlock(&obj->inner_lock);

	return obj->has_lock;
}

void
PeriodLocker_ReserveReleaseLock(period_locker_t locker)
{
	period_locker* obj = (period_locker*)locker;

	(void)pthread_mutex_lock(&obj->inner_lock);
	if (! obj->has_lock) {
		fprintf(stderr, "ERR! the lock is already unlocked.\n");
		(void)pthread_mutex_unlock(&obj->inner_lock);
		return;
	}
	obj->release_rtime = GetCurrentTimeInMillis();
	obj->is_release_reserved = true;
	(void)pthread_mutex_unlock(&obj->inner_lock);
}


//#define ENABLE_UNIT_TEST	1
#define ENABLE_UNIT_TEST	0
#if ENABLE_UNIT_TEST
int
main(int argc, char** argv)
{
	period_locker_t	locker = PeriodLocker_Create();

	if (PeriodLocker_AcquireLock(locker)) {
		printf("lock is acquired.\n");
	} else {
		printf("cannot acquire the lock!\n");
		PeriodLocker_Destroy(locker, NULL, NULL);
		return -1;
	}
	PeriodLocker_ReserveReleaseLock(locker);
//	printf("sleep short time after release...\n");
	(void)sleep(1/*2*/);
	if (PeriodLocker_AcquireLock(locker)) {
		printf("lock is acquired again.\n");
	}
	(void)sleep(/*1*/6);
//	printf("sleep a while after release...\n");
	PeriodLocker_ReserveReleaseLock(locker);
//	printf("sleep enough time after release...\n");
	(void)sleep(30);
	if (PeriodLocker_AcquireLock(locker)) {
		printf("lock is acquired again.\n");
	}
	printf("sleep enough time while holding lock...\n");
	(void)sleep(30);
	PeriodLocker_Destroy(locker, NULL, NULL);

	return 0;
}
#endif  // ENABLE_UNIT_TEST

//
// End of File
//
